# -*- coding: UTF-8 -*-
# Python
"""
11-09-2023
@author: jeremy auclair

Contains functions to facilitate output plotting and processing
"""

import numpy as np  # array data
from typing import List  # to declare variables
from matplotlib import colors  # for colormap creation


def hex_to_rgb(value: str) -> List[int]:
    """
    Converts hex to rgb colours
    
    Arguments
    =========

    1. value: ``str``
        string of 6 characters representing a hex colour
    
    Returns
    =======

    1. rbg: ``List[int]``
        list length 3 of RGB values
    """
    value = value.strip("#") # removes hash symbol if present
    len_value = len(value)
    rgb = tuple(int(value[i:i + len_value // 3], 16) for i in range(0, len_value, len_value // 3))
    return rgb


def rgb_to_dec(value: List[int]) -> List[float]:
    """
    Converts rgb to decimal colours (i.e. divides each value by 256)

    Arguments
    =========

    1. value: ``List[int]``
        list (length 3) of RGB values

    Returns
    =======
    
    1. dec: ``List[float]``
        list (length 3) of decimal values
    """
    dec = [v/256 for v in value]
    return dec


def get_continuous_cmap(hex_list: List[str], float_list: List[float] = None) -> colors.LinearSegmentedColormap:
    """
    Creates and returns a color map that can be used in heat map figures.
    If float_list is not provided, colour map graduates linearly between each color in hex_list.
    If float_list is provided, each color in hex_list is mapped to the respective location in float_list. 
        
    Arguments
    =========

    1. hex_list: ``List[str]``
        list of hex code strings
    2. float_list: ``List[float]``
        list of floats between 0 and 1, same length as hex_list. Must start with 0 and end with 1.
        
    Returns
    =======

    1. cmp: ``matplotlib.colors.LinearSegmentedColormap``
        converted colour map
    """
    rgb_list = [rgb_to_dec(hex_to_rgb(i)) for i in hex_list]
    if float_list:
        pass
    else:
        float_list = list(np.linspace(0,1,len(rgb_list)))
        
    cdict = dict()
    for num, col in enumerate(['red', 'green', 'blue']):
        col_list = [[float_list[i], rgb_list[i][num], rgb_list[i][num]] for i in range(len(float_list))]
        cdict[col] = col_list
    cmp = colors.LinearSegmentedColormap('my_cmp', segmentdata=cdict, N=256)
    return cmp