# -*- coding: UTF-8 -*-
# Python
"""
18-09-2023
@author: jeremy auclair

Convert netCDF4 rasters to dataframes and geodataframes for the parcel mode.
"""

import os  # for path management
from typing import Union  # to declare variables
import xarray as xr  # to manage nc files
import pandas as pd  # to manage dataframes
import geopandas as gpd  # to manage shapefile crs projections
from typing import List  # to declare variables


def convert_xarray_to_dataframe(dataset_path: Union[str, xr.Dataset, xr.DataArray], save_path: str, shapefile_path: str = None, output_shapefile_dir: str = None, shapefile_variables: List[str] = [], time_dimension: bool = True, add_ETR = False) -> None:
    """
    Convert the modspa output netCDF dataset back to a ``pd.DataFrame``
    and save it as a csv file for the parcel mode.

    Arguments
    =========

    1. dataset_path: ``Union[str, xr.Dataset, xr.DataArray]``
        path to the netCDF file
    2. save_path: ``str``
        path to save the ``pd.DataFrame`` as a csv file
    3. shapefile_path: ``str`` ``default = None``
        path to original shapefile
    4. output_shapefile_dir: ``str`` ``default = None``
        path to directory to save new shapefiles
    5. shapefile_variables: ``List[str]`` ``default = []``
        list of variables to save as shapefiles. One shapefile
        per variable will be saved
    6. time_dimension: ``bool`` ``default = True``
        boolean to indicate if the dataframe has a time dimension
    7. add_ETR: ``bool`` ``default = False``
        boolean to choose to add the ETR (E + Tr) column to the 
        dataframe

    Returns
    =======

    ``None``
    """
    
    # Create output shapefile directory if it is not None and is does not exist
    if output_shapefile_dir:
        if not os.path.exists(output_shapefile_dir): os.mkdir(output_shapefile_dir)
    
    # If dataframe has a time dimenson (multiple dates)
    if time_dimension:
        
        # Open dataset
        if type(dataset_path) == str:
            dataset = xr.open_dataset(dataset_path, chunks = 'auto').squeeze(dim = 'y', drop = True)  # remove unecessary y dimension
        else:
            dataset = dataset_path.squeeze(dim = 'y', drop = True)  # remove unecessary y dimension

        # Convert to dataframe and rename indexes
        dataframe = dataset.to_dataframe(dim_order = ['time', 'x'])
        dataframe.index.names = ['date','id']
        
        # Close xarray dataset
        dataset.close()
        
        # Remove multi index
        dataframe = dataframe.reset_index(level=[1])
        dataframe['date'] = dataframe.index.values
        
        if add_ETR:
            dataframe['ETR'] = dataframe['E'] + dataframe['Tr']
        
        # Save dataframe to csv
        dataframe.to_csv(save_path, index = False)
        
        # Merge dataframe with shapefile
        if shapefile_path:
            
            # Open shapefile
            shapefile = gpd.read_file(shapefile_path)
            
            # Add index column
            shapefile['id'] = shapefile.index + 1
            
            for var in shapefile_variables:
                
                # Create save path
                shapefile_save_path = output_shapefile_dir + os.sep + var + '_shapefile'
                
                print(f'Merging csv and shapefile: {shapefile_save_path}')
                
                # Pivot dataframe with only current variable
                var_dataframe = dataframe.pivot(index = 'id', columns = 'date', values = var).reset_index()
            
                # Merge shapefile and dataframe
                new_shapefile = shapefile.merge(var_dataframe, left_on = 'id', right_on = 'id')
                
                # Convert pandas timestamps to string
                new_shapefile = new_shapefile.rename(columns = lambda x: x.strftime('%Y-%m-%d') if isinstance(x, pd.Timestamp) else x)
                
                # Save shapefile
                new_shapefile.to_file(shapefile_save_path, index = False)
            
    else:
        
        # Open dataset
        if type(dataset_path) == str:
            dataset = xr.open_dataset(dataset_path, chunks = 'auto').squeeze(dim = 'y', drop = True)  # remove unecessary y dimension
        else:
            dataset = dataset_path.squeeze(dim = 'y', drop = True)  # remove unecessary y dimension

        # Convert to dataframe and rename indexes
        dataframe = dataset.to_dataframe(dim_order = ['x'])
        dataframe.index.name = ['id']
        
        # Close xarray dataset
        dataset.close()
        
        if add_ETR:
            dataframe['ETR'] = dataframe['E'] + dataframe['Tr']
        
        # Save dataframe to csv
        dataframe.to_csv(save_path, index = False)
        
        # Merge dataframe with shapefile
        if shapefile_path:
            
            # Open shapefile
            shapefile = gpd.read_file(shapefile_path)
            
            # Add index column
            shapefile['id'] = shapefile.index + 1
            
            # Merge shapefile and dataframe
            new_shapefile = shapefile.merge(dataframe, left_on = 'id', right_on = 'id')
            new_shapefile = new_shapefile.set_index('id')
            
            # Create save path
            shapefile_save_path = output_shapefile_dir + os.sep + 'multivariable_shapefile'
            
            # Save shapefile
            new_shapefile.to_file(shapefile_save_path, index = True)
        
    return None