# -*- coding: UTF-8 -*-
# Python
"""
08-09-2023
@author: jeremy auclair

Main script to plot ETR and NDVI GIFs
"""

if __name__ == '__main__':
    
    import sys, os  # system management
    currentdir = os.path.dirname(os.path.realpath(__file__))
    parentdir = os.path.dirname(currentdir)
    sys.path.insert(0, os.path.dirname(parentdir))
    
    
    import argparse  # for parsing arguments
    import geopandas as gpd  # to manage shapefiles
    import pandas as pd  # to manage dataframes
    import numpy as np  # vectorized math
    import xarray as xr  # to manage dataset
    import json  # to load scaling dictionnary
    import matplotlib.pyplot as plt  # to plot
    from matplotlib import rcParams  # plot parameters
    from mpl_toolkits.axes_grid1 import make_axes_locatable  # for pyplot colorbar
    import contextily as cx  # to add basemaps
    import xyzservices.providers as xyz  # map providers
    import warnings  # to turn off warnings
    from p_tqdm import p_umap  # for multiprocessing with progress bars
    from PIL import Image  # to generate GIF
    import contextlib  # to generate GIF
    import gc  # garbage collector module
    import matplotlib  # to prevent showing graphs
    matplotlib.use('Agg')  # to prevent showing graphs
    from modspa_pixel.config.config import config  # to import config file
    from modspa_pixel.postprocessing.output_toolbox import get_continuous_cmap
    from modspa_pixel.source.code_toolbox import format_duration  # to print memory requirements
    from time import time  # to time code excecution
    
    # Get start time of script
    t0 = time()
    
    ## Arguments to run the script
    parser = argparse.ArgumentParser()
    parser.add_argument("variable", type = str, choices = ['NDVI', 'ETR', 'DP', 'Irr', 'E', 'Tr', 'SWCe', 'SWCr'],  help = "Variables to generate a daily GIF from, choices = NDVI, ETR, DP, Irr, E, Tr, SWCe, SWCr. For example, to plot a NDVI GIF, use this command: python plot_parcel_GIF.py NDVI")
    try:
        args = parser.parse_args()
    except:
        print("\nVariables to generate a daily GIF from, choices = NDVI, ETR, DP, Irr, E, Tr, SWCe, SWCr.\n\nFor example, to plot a NDVI GIF, use this command: python plot_parcel_GIF.py NDVI\n")
        sys.exit(0)
    variable = args.variable

    # Turn off warnings
    np.seterr(divide='ignore', invalid='ignore')
    warnings.simplefilter("ignore", category = RuntimeWarning)

    # Parameters for matplotlib
    plt.style.use('default')
    rcParams['mathtext.fontset'] = 'stix'
    rcParams['font.family'] = 'STIXGeneral'
    rcParams.update({'font.size': 15})
    
    # Custom colormap for ndvi
    hex_list = ['ffffff', 'ce7e45', 'df923d', 'f1b555', 'fcd163', '99b718',
                '74a901', '66a000', '529400', '3e8601', '207401', '056201',
                '004c00', '023b01', '012e01', '011d01', '011301']
    cmap_ndvi = get_continuous_cmap(hex_list)
    
    # Declare paths
    config_file = parentdir + os.sep + 'config' + os.sep + 'config_modspa.json'
    
    # Open config file and load parameters
    config_params = config(config_file)
    
    # General parameters
    max_cpu = config_params.max_cpu
    run_name = config_params.run_name
    preferred_provider = config_params.preferred_provider
    start_date = config_params.start_date
    end_date = config_params.end_date

    # Load paths
    output_path = config_params.output_path
    download_path = config_params.download_path
    
    # Shapefile path
    shapefile_path = config_params.shapefile_path
    
    # Output path
    output_save_path = output_path + os.sep + run_name + os.sep + run_name + '_outputs.nc'
    output_dataframe_path = output_path + os.sep + run_name + os.sep + run_name + '_outputs.csv'
    
    # NDVI input path
    if preferred_provider == 'copernicus':
        ndvi_path = download_path + os.sep + 'SCIHUB' + os.sep + 'NDVI' + os.sep + run_name + os.sep + 'ndvi_interp.csv'
    elif preferred_provider == 'theia':
        ndvi_path = download_path + os.sep + 'THEIA' + os.sep + 'NDVI' + os.sep + run_name + os.sep + 'ndvi_interp.csv'
    elif preferred_provider == 'usgs':
        ndvi_path = download_path + os.sep + 'USGS' + os.sep + 'NDVI' + os.sep + run_name + os.sep + 'ndvi_interp.csv'

    # Convert output netCDF to pandas dataframe
    output_dataset = xr.open_dataset(output_save_path)
    scaling = json.loads(output_dataset.attrs['scaling'].replace("'", '"'))
    output_dataset.close()
    
    # Open output dataframe
    output_dataframe = pd.read_csv(output_dataframe_path)
    dates = list(dict.fromkeys(output_dataframe['date'].values))
    
    # Open shapefile
    shapefile = gpd.read_file(shapefile_path)
    
    # Save path
    output_gif_path = output_path  + os.sep + run_name + os.sep + 'GIF'
    if not os.path.exists(output_gif_path): os.mkdir(output_gif_path)
    gif_var_path = output_gif_path + os.sep + variable + '.gif'
    
    # Define plot function
    def plot_image(args: tuple) -> str:
        """
        Generate a matplotlib image and save it. Used as a frame for
        GIF generation.

        Arguments (packed in args: ``tuple``)
        =====================================

        1. day: ``int``
            index of current day
        2. shapefile_plot: ``gpd.GeoDataFrame``
            DataArray of current day
        3. plot_data: ``pd.Series``
            data to plot current day
        4. cmap: ``matplotlib.colors.Colormap``
            colormap to plot
        4. title: ``str``
            title to show for current figure
        5. variable: ``str``
            name of variable (for save file)
        6. colorbar_unit: ``str``
            colorbar unit for label
        6. vmin: ``float``
            minimum value of data
        7. vmax: ``float``
            maximum value of data
        8. output_gif_path: ``float``
            directory in which to save images

        Returns
        =======

        1. figure_path : ``str``
            path of saved image
        """
        
        # Unpack args
        day, shapefile_plot, plot_data, cmap, title, variable, colorbar_label, vmin, vmax, output_gif_path = args
        
        # Add data to shapefile
        shapefile_plot[variable] = plot_data.values
        
        # Prepare figure
        fig, ax = plt.subplots(1, figsize=(12, 6))

        # Set ticks for grid
        xmin, ymin, xmax, ymax = shapefile_plot.total_bounds
        major_ticks_x = np.linspace(xmin, xmax, 5, endpoint = True)
        minor_ticks_x = np.linspace(xmin, xmax, 21, endpoint = True)
        major_ticks_y = np.linspace(ymin, ymax, 5, endpoint = True)
        minor_ticks_y = np.linspace(ymin, ymax, 21, endpoint = True)
        x_text_ticks = [np.format_float_scientific(t, precision = 2, exp_digits = 1, trim = '0') for t in major_ticks_x]
        y_text_ticks = [np.format_float_scientific(t, precision = 2, exp_digits = 1, trim = '0') for t in major_ticks_y]
        ax.set_xticks(major_ticks_x)
        ax.set_xticklabels(x_text_ticks)
        ax.set_xticks(minor_ticks_x, minor = True)
        ax.set_yticks(major_ticks_y, labels = y_text_ticks)
        ax.set_yticklabels(y_text_ticks)
        ax.set_yticks(minor_ticks_y, minor = True)
        ax.grid(which = 'minor', alpha = 0.2, zorder = 3, linewidth = 0.5, color = 'dimgray')
        ax.grid(which = 'major', alpha = 0.5, zorder = 3, linewidth = 1, color = 'dimgray')

        # Plot data
        shapefile_plot.plot(column = variable, ax = ax, cmap = cmap, vmin = vmin, vmax = vmax, zorder = 10)
        ax.set_aspect('equal', 'box')

        # Set title and axis labels
        xrange = abs(xmax - xmin)
        yrange = abs(ymax - ymin)
        plt.title(title);
        plt.xlabel('$x \: [m]$')
        plt.ylabel('$y \: [m]$')
        ax.set_xlim(xmin - 0.01*xrange, xmax + 0.01*xrange)
        ax.set_ylim(ymin - 0.01*yrange, ymax + 0.01*yrange)

        # Set an axis for the color bar
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size = "4%", pad = 0.2)

        # Add and manage color bar
        mappable = plt.cm.ScalarMappable(cmap = cmap, norm = plt.Normalize(vmin = vmin, vmax = vmax))
        cbar = fig.colorbar(mappable, cax=cax)
        cbar.set_label(label =  colorbar_label, labelpad = 15, y = 0.5)
        fig.tight_layout()

        # Add basemap
        cx.add_basemap(ax, crs = shapefile_plot.crs, source = xyz.OpenTopoMap, alpha = 0.6, reset_extent = True, attribution = False)
        
        # Create save_path and save
        figure_path = output_gif_path + os.sep + f'{day+1:03d}_' + variable + '_.png'
        fig.savefig(figure_path, dpi = 100, bbox_inches='tight')
        
        # Release memory
        plt.close()
        fig.clf()
        del fig, ax  # delete the objects
        gc.collect()  # collect garbage
        
        return figure_path

    
    if variable == 'ETR':
    
        # Generate arguments for ETR GIF
        vmin = 0.1
        vmax = 6
        
        # Prepare shapefile for plot
        shapefile_plot = shapefile.copy()
        
        args = [(day, shapefile_plot, output_dataframe[output_dataframe['date'] == dates[day]][variable] / scaling['E'], 'YlGnBu', 'Evapotranspiration -- '  + dates[day], variable, 'E + Tr [mm]', vmin, vmax, output_gif_path) for day in range(len(dates))]

        # Generate frames
        print('\nGenerating frames for ETR GIF\n')
        results = p_umap(plot_image, args, **{"num_cpus": max_cpu})
        results.sort()
        
        # Generate GIF for ETR
        print('\nGenerating ETR GIF\n')
        with contextlib.ExitStack() as stack:

            # lazily load images
            imgs = (stack.enter_context(Image.open(f)) for f in results)

            # extract  first image from iterator
            img = next(imgs)

            # Save GIF
            img.save(fp = gif_var_path, format = 'GIF', append_images = imgs, save_all = True, duration = 100, loop = 2)

        # Remove unecessary images
        for file in results:
            os.remove(file)
    
    elif variable == 'NDVI':
        
        # Open NDVI dataset
        ndvi_dataframe = pd.read_csv(ndvi_path)
            
        # Generate arguments for NDVI GIF
        vmin = 0
        vmax = 1
        
        # Prepare shapefile for plot
        shapefile_plot = shapefile.copy()
        
        args = [(day, shapefile_plot, ndvi_dataframe[ndvi_dataframe['date'] == dates[day]][variable] / 255, cmap_ndvi, 'NDVI -- '  + dates[day], variable, 'NDVI', vmin, vmax, output_gif_path) for day in range(len(dates))]

        # Generate frames
        print('\nGenerating frames for NDVI GIF\n')
        results = p_umap(plot_image, args, **{"num_cpus": max_cpu})
        results.sort()
        
        # Generate GIF for NDVI
        print('\nGenerating NDVI GIF\n')
        with contextlib.ExitStack() as stack:

            # lazily load images
            imgs = (stack.enter_context(Image.open(f)) for f in results)

            # extract  first image from iterator
            img = next(imgs)
            
            # Save GIF
            img.save(fp = gif_var_path, format = 'GIF', append_images = imgs, save_all = True, duration = 100, loop = 2)

        # Remove unecessary images
        for file in results:
            os.remove(file)
    
    elif variable in ['E', 'Tr', 'SWCe', 'SWCr']:
        
        # Generate arguments for variable GIF
        if variable in ['E', 'Tr']:
            vmin = 0.1
            vmax = 4
            unit = ' [mm]'
            if variable == 'E': variable_name = 'Evaporation'
            else: variable_name = 'Transpiration'
        else:
            vmin = 0
            vmax = 1
            unit = ''
            if variable == 'SWCe': variable_name = 'Soil Water Content evaporative layer'
            else: variable_name = 'Soil Water Content root layer'
        
        # Prepare shapefile for plot
        shapefile_plot = shapefile.copy()
        
        args = [(day, shapefile_plot, output_dataframe[output_dataframe['date'] == dates[day]][variable] / scaling[variable], 'YlGnBu', variable_name + ' -- '  + dates[day], variable, variable + unit, vmin, vmax, output_gif_path) for day in range(len(dates))]

        # Generate frames
        print(f'\nGenerating frames for {variable} GIF\n')
        results = p_umap(plot_image, args, **{"num_cpus": max_cpu})
        results.sort()
        
        # Generate GIF for ETR
        print(f'\nGenerating {variable} GIF\n')
        with contextlib.ExitStack() as stack:

            # lazily load images
            imgs = (stack.enter_context(Image.open(f)) for f in results)

            # extract  first image from iterator
            img = next(imgs)

            # Save GIF
            img.save(fp = gif_var_path, format = 'GIF', append_images = imgs, save_all = True, duration = 100, loop = 2)

        # Remove unecessary images
        for file in results:
            os.remove(file)
    
    else:
        
        
        # Generate arguments for variable GIF
        if variable == 'DP': variable_name = 'Deep Percolation'
        else: variable_name = 'Irrigation'
        vmin = output_dataframe[variable].min().values
        vmax = output_dataframe[variable].max().values
        
        # Prepare shapefile for plot
        shapefile_plot = shapefile.copy()
        
        args = [(day, shapefile_plot, output_dataframe[output_dataframe['date'] == dates[day]][variable] / scaling[variable], 'YlGnBu', variable_name + ' -- '  + dates[day], variable, variable + ' [mm]', vmin, vmax, output_gif_path) for day in range(len(dates))]

        # Generate frames
        print(f'\nGenerating frames for {variable} GIF\n')
        results = p_umap(plot_image, args, **{"num_cpus": max_cpu})
        results.sort()
        
        # Generate GIF for ETR
        print(f'\nGenerating {variable} GIF\n')
        with contextlib.ExitStack() as stack:

            # lazily load images
            imgs = (stack.enter_context(Image.open(f)) for f in results)

            # extract  first image from iterator
            img = next(imgs)

            # Save GIF
            img.save(fp = gif_var_path, format = 'GIF', append_images = imgs, save_all = True, duration = 100, loop = 2)

        # Remove unecessary images
        for file in results:
            os.remove(file)
    
    # Print formatted runtime
    format_duration(time() - t0)